from datetime import datetime
import os
import shutil

from flask import current_app
from whoosh import writing
from whoosh.index import create_in, open_dir

from . import zotero
from .extractors import ItemContext, LibraryContext


def open_index(auto_create=False):
    if not current_app.config['KERKO_DATA_DIR'].exists() and auto_create:
        os.makedirs(current_app.config['KERKO_DATA_DIR'], exist_ok=True)
        return create_in(
            current_app.config['KERKO_DATA_DIR'],
            current_app.config['KERKO_COMPOSER'].schema
        )
    if current_app.config['KERKO_DATA_DIR'].exists():
        return open_dir(current_app.config['KERKO_DATA_DIR'])
    raise FileNotFoundError


def delete_index():
    if current_app.config['KERKO_DATA_DIR'].exists():
        shutil.rmtree(current_app.config['KERKO_DATA_DIR'])


def request_library_context(zotero_credentials):
    collections = zotero.Collections(zotero_credentials)
    item_types = {
        t['itemType']: t['localized']
        for t in zotero.load_item_types(zotero_credentials)
    }
    item_fields = {
        t: zotero.load_item_type_fields(zotero_credentials, t)
        for t in item_types.keys()
    }
    creator_types = {
        t: zotero.load_item_type_creator_types(zotero_credentials, t)
        for t in item_types.keys()
    }
    return LibraryContext(collections, item_types, item_fields, creator_types)


def update_index():
    start_time = datetime.now()

    composer = current_app.config['KERKO_COMPOSER']
    zotero_credentials = zotero.init_zotero()
    library_context = request_library_context(zotero_credentials)
    index = open_index(auto_create=True)
    count = 0
    writer = index.writer(limitmb=256)
    try:
        writer.mergetype = writing.CLEAR
        allowed_item_types = [
            t for t in library_context.item_types.keys()
            if t not in ['note', 'attachment']
        ]
        for item, notes in zotero.Items(zotero_credentials, allowed_item_types):
            count += 1
            document = {}
            item_context = ItemContext(item, notes)
            for spec in list(composer.fields.values()) + list(composer.facets.values()):
                spec.extract_to_document(document, item_context, library_context)
            update_document_with_writer(writer, document, count=count)
    except Exception as e:  # pylint: disable=broad-except
        writer.cancel()
        current_app.logger.exception(e)
        current_app.logger.error('An exception occurred. Could not finish updating the index.')
    else:
        writer.commit()
        current_app.logger.info(_format_elapsed_time(count, start_time))


def update_document_with_writer(writer, document, count=None):
    """
    Update a document in the search index.

    :param writer: The index writer.

    :param document: A dict whose fields match the schema.

    :param count: An optional document count, for logging purposes.
    """
    writer.update_document(**document)
    current_app.logger.debug(
        'Document {count}updated ({id}): {title}'.format(
            id=document.get('id', ''),
            title=document.get('z_title'),
            count='' if count is None else '{} '.format(count)
        )
    )


def update_document(document):
    """
    Update a document in the search index.

    :param document: A dict whose fields match the schema.
    """
    index = open_index()
    with index.writer(limitmb=256) as writer:
        update_document_with_writer(writer, document)


def _format_elapsed_time(count, start_time):
    elapsed_time = int(round((datetime.now() - start_time).total_seconds()))
    elapsed_min, elapsed_sec = elapsed_time // 60, elapsed_time % 60
    s = ('{num} items indexed in' if count else '{num} items indexed in').format(num=count)
    if elapsed_min > 0:
        s += (' {num} minute' if elapsed_min else ' {num} minutes').format(num=elapsed_min)
        s += (' {num:02} second' if elapsed_sec else ' {num:02d} seconds').format(num=elapsed_sec)
    else:
        s += (' {num} second' if elapsed_sec else ' {num} seconds').format(num=elapsed_sec)
    return s



from flask import current_app
from whoosh.qparser import MultifieldParser, QueryParser
from whoosh.qparser.plugins import PhrasePlugin, GroupPlugin, OperatorsPlugin
from whoosh.query import And, Not, Every
from whoosh.sorting import Facets, Count

from .criteria import Criteria
from .index import open_index


def build_keywords_query(keywords):
    """
    Build parsers for a query.

    :param MultiDict keywords: The search texts keyed by scope key. If empty,
        the query will match every documents.
    """
    queries = []
    if keywords:
        composer = current_app.config['KERKO_COMPOSER']
        text_plugins = [PhrasePlugin(),
                        GroupPlugin(),
                        OperatorsPlugin()]
        for key, value in keywords.items(multi=True):
            fields = [spec.key for spec in composer.fields.values() if key in spec.scopes]
            if not fields:
                raise KeyError  # No known field for that scope key.
            parser = MultifieldParser(
                fields, schema=composer.schema, plugins=text_plugins
            )
            queries.append(parser.parse(value))
    else:
        queries.append(Every())
    return And(queries)


def build_filter_query(filters=None):
    """
    Build groupedby and filter queries based on facet specs.

    :param list filter: A list of (name, values) tuples, where values is itself
        a list.

    :return: A tuple with the Facets to perform grouping on, and the terms to
        filter on.
    """
    composer = current_app.config['KERKO_COMPOSER']
    groupedby = Facets()
    for spec in composer.facets.values():
        groupedby.add_field(spec.key, allow_overlap=spec.allow_overlap)

    terms = []
    if filters:
        for filter_key, filter_values in filters:
            spec = composer.get_facet_by_filter_key(filter_key)
            if spec:  # Ensure only valid filters.
                for v in filter_values:
                    if v == '':  # If trying to filter with a missing value.
                        # Exclude all results with a value in facet field.
                        terms.append(Not(Every(spec.key)))
                    else:
                        v = spec.codec.transform_for_query(v)
                        terms.append(spec.query_class(spec.key, v))
    return groupedby, And(terms)


def _get_fields(hit, return_fields=None):
    """
    Copy fields from a Hit object into a dict.

    Fields must be copied while the Searcher object is alive.

    :param whoosh.searching.Hit hit: The search result.

    :param list return_fields: A list of fields to load. If None, all fields
        will be loaded.
    """
    if return_fields is None:
        item = hit.fields()
    else:
        item = {f: hit[f] for f in return_fields if f in hit}
    return _decode_fields(item)


def _decode_fields(item):
    field_specs = current_app.config['KERKO_COMPOSER'].fields
    for k, v in item.items():
        if k in field_specs:
            item[k] = field_specs[k].decode(v)
    return item


def build_creators_display(item):
    """Prepare creator data for easier display."""
    for spec in item.keys():
        if spec == 'data' and 'creators' in item['data']:
            for creator in item['data']['creators']:
                # Add creator display name.
                creator['display'] = creator.get('name', '') or ', '.join(
                    [
                        n for n in (
                            creator.get('lastName', ''),
                            creator.get('firstName', '')
                        ) if n
                    ]
                )
                # Add creator type labels.
                if 'creator_types' in item:
                    for t in item['creator_types']:
                        if t['creatorType'] == creator['creatorType']:
                            creator['label'] = t['localized']
                            break
                # Add creator link.
                if 'creator' in current_app.config['KERKO_COMPOSER'].scopes:
                    creator['url'] = Criteria().build_add_keywords_url(
                        scope='creator',
                        value='"{}"'.format(creator['display'])
                    )


def build_fake_facet_results(item):
    """
    Prepare facet "results" describing the item.

    The only distinction from real facet results obtained from a search is that
    counts are unknown. Using the same data structure as real facet results
    allows the reuse of facet display logic.
    """
    item['facet_results'] = {}
    for spec in current_app.config['KERKO_COMPOSER'].facets.values():
        if spec.key in item:
            if isinstance(item[spec.key], list):
                fake_results = [(value, 0) for value in item[spec.key]]
            else:
                fake_results = [(item[spec.key], 0)]
            # Pass an empty Criteria; the facets will provide starting points for new searches.
            item['facet_results'][spec.key] = spec.build(fake_results, criteria=Criteria())


def run_query(criteria, return_fields=None):
    """Perform a search query."""
    items = []
    facets = {}
    total = 0
    page_count = 0

    index = open_index()
    if index:
        with index.searcher() as searcher:
            composer = current_app.config['KERKO_COMPOSER']
            q = build_keywords_query(criteria.keywords)
            g, fq = build_filter_query(criteria.filters.lists())
            search_args = {
                'groupedby': g,
                'filter': fq,
                'maptype': Count,
                'sortedby': composer.sorts[criteria.sort].get_field_keys(),
                'reverse': composer.sorts[criteria.sort].reverse,
            }

            if criteria.page_len is None:  # Retrieve all results.
                search_args['limit'] = None

                results = searcher.search(q, **search_args)
                if results:
                    items = [_get_fields(hit, return_fields) for hit in results]
                    for spec in composer.facets.values():
                        facets[spec.key] = spec.build(
                            results.groups(spec.key).items(), criteria
                        )
                    total = results.estimated_length()
                    page_count = 1
            else:  # Retrieve a range of results.
                search_args['pagenum'] = criteria.page_num
                search_args['pagelen'] = criteria.page_len

                results = searcher.search_page(q, **search_args)
                if results:
                    items = [_get_fields(hit, return_fields) for hit in results]
                    for spec in composer.facets.values():
                        facets[spec.key] = spec.build(
                            results.results.groups(spec.key).items(), criteria
                        )
                    total = results.total
                    page_count = results.pagecount

    return items, facets, total, page_count


def run_query_unique(field_name, value, return_fields=None):
    """Perform a search query for a single item using an unique key."""
    index = open_index()
    if index:
        with index.searcher() as searcher:
            q = QueryParser(
                field_name,
                schema=current_app.config['KERKO_COMPOSER'].schema,
                plugins=[]
            ).parse(value)
            results = searcher.search(q, limit=1)
            if results:
                return _get_fields(results[0], return_fields)
    return None



import pprint

import click
from flask.cli import with_appcontext

from . import zotero
from .index import update_index, delete_index


@click.group()
def cli():
    """Run a Kerko subcommand."""


@cli.command()
@with_appcontext
def index():
    """Index items from the Zotero library."""
    update_index()


@cli.command()
@with_appcontext
def clean():
    """Delete the search index data immediately."""
    delete_index()


@cli.command()
@click.argument('item_key')
@with_appcontext
def zotero_item(item_key):
    """
    Retrieve an item from the Zotero library.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item(credentials, item_key))


@cli.command()
@with_appcontext
def zotero_item_types():
    """
    List all item types supported by Zotero.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_types(credentials))


@cli.command()
@with_appcontext
def zotero_item_fields():
    """
    List all fields supported by Zotero.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_fields(credentials))


@cli.command()
@click.argument('item_type')
@with_appcontext
def zotero_item_type_fields(item_type):
    """
    List all fields supported by Zotero for a given item type.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_type_fields(credentials, item_type))


@cli.command()
@click.argument('item_type')
@with_appcontext
def zotero_item_type_creator_types(item_type):
    """
    List all creator types supported by Zotero for a given item type.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    pprint.pprint(zotero.load_item_type_creator_types(credentials, item_type))


@cli.command()
@with_appcontext
def zotero_top_level_collections():
    """
    List keys and names of top-level collections from the Zotero library.

    WARNING: This command is provided for debugging purposes only and may be
    modified or removed from the module at any time.
    """
    credentials = zotero.init_zotero()
    collections = zotero.Collections(credentials, top_level=True)
    for c in collections:
        print(f"{c.get('key')} {c.get('data', {}).get('name', '')}")



from time import sleep
import requests

import wrapt

from flask import current_app
from pyzotero import zotero, zotero_errors


@wrapt.decorator
def retry_zotero(wrapped, _instance, args, kwargs):
    """
    Retry the wrapped function if the Zotero API call fails.

    Caution: This decorator should only be used on simple functions, as the
    whole function is called repeatedly as long a Zotero fails.
    """
    attempts = 1
    while True:
        try:
            return wrapped(*args, **kwargs)
        except (
                requests.exceptions.ConnectionError, zotero_errors.HTTPError
        ) as e:
            current_app.logger.exception(e)
            if attempts < current_app.config['KERKO_ZOTERO_MAX_ATTEMPTS']:
                current_app.logger.warning(
                    "The Zotero API call has failed in {func}. "
                    "New attempt in {wait} seconds...".format(
                        func=wrapped.__name__,
                        wait=current_app.config['KERKO_ZOTERO_WAIT']
                    )
                )
                attempts += 1
                sleep(current_app.config['KERKO_ZOTERO_WAIT'])
            else:
                current_app.logger.error(
                    "The maximum number of API call attempts to Zotero has "
                    "been reached. Stopping."
                )
                raise


def init_zotero():
    return zotero.Zotero(
        library_id=current_app.config['KERKO_ZOTERO_LIBRARY_ID'],
        library_type=current_app.config['KERKO_ZOTERO_LIBRARY_TYPE'],
        api_key=current_app.config['KERKO_ZOTERO_API_KEY'],
        locale=current_app.config['KERKO_ZOTERO_LOCALE']
    )


@retry_zotero
def load_item(zotero_credentials, item_key):
    """Return a specific item."""
    return zotero_credentials.item(item_key)


@retry_zotero
def load_item_types(zotero_credentials):
    """Return all Zotero item types."""
    return zotero_credentials.item_types()


@retry_zotero
def load_item_fields(zotero_credentials):
    """Return all Zotero fields."""
    return zotero_credentials.item_fields()


@retry_zotero
def load_item_type_fields(zotero_credentials, item_type):
    """Return all Zotero fields for a given item type."""
    return zotero_credentials.item_type_fields(item_type)


@retry_zotero
def load_item_type_creator_types(zotero_credentials, item_type):
    """List all Zotero creator types for a given item type."""
    return zotero_credentials.item_creator_types(item_type)


class Collections:
    """
    Iterable over all collections found in Zotero.

    Collections are loaded in memory all at once, and may be acccessed by key
    using the dict[] notation.

    Each collection is a dict as returned by Zotero. Example::

        {
            'data': {
                'key': 'C435CGN4',
                'name': 'Example Collection Name',
                'parentCollection': 'D5U2TEVB',
                'relations': {   },
                'version': 114
            },
            'key': 'C435CGN4',
            'library': {
                'id': 2076305,
                'links': {
                    'alternate': {
                        'href': 'https://www.example.com',
                        'type': 'text/html'
                    }
                },
                'name': 'Example Library',
                'type': 'user'
            },
            'links': {
                'alternate': {
                    'href': 'https://www.zotero.org/example_library/collections/C435CGN4',
                    'type': 'text/html'
                },
                'self': {
                    'href': 'https://api.zotero.org/users/9999999/collections/C435CGN4',
                    'type': 'application/json'
                },
                'up': {
                    'href': 'https://api.zotero.org/users/9999999/collections/D5U2TEVB',
                    'type': 'application/atom+xml'
                }
            },
            'meta': {
                'numCollections': 0,
                'numItems': 15
            },
            'version': 114
        }
    """

    def __init__(self, zotero_credentials, top_level=False):
        self.collections = {}
        # Immediately load all collections, to allow later access by key.
        self.load_collections(zotero_credentials, top_level)
        self.iterator = None

    @retry_zotero
    def load_collections(self, zotero_credentials, top_level):
        start = 0
        if top_level:
            method = zotero_credentials.collections_top
        else:
            method = zotero_credentials.collections
        while True:
            more = method(start=start, limit=current_app.config['KERKO_ZOTERO_BATCH_SIZE'])
            if not more:
                break
            start += len(more)
            for collection in more:
                self.collections[collection['key']] = collection

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.collections)

    def __getitem__(self, key):
        return self.collections[key]

    def __contains__(self, key):
        return key in self.collections

    def __next__(self):
        if self.iterator is None:
            self.iterator = iter(self.collections.values())
        return next(self.iterator)

    def get(self, key, default):
        return self.collections.get(key, default)

    def ancestors(self, key):
        """
        Return the ancestors of the specified collection.

        :return: A list representing the path of the specified collection,
            including the specified collection. Each list element is a
            collection key.
        """
        collection = self.collections.get(key)
        if collection:
            parent_key = collection['data'].get('parentCollection')
            if parent_key:
                ancestors = self.ancestors(parent_key)
                ancestors.append(key)
                return ancestors
            return [key]
        return []


class Items:
    """
    Iterable over Zotero items.

    Items are loaded on demand, in small batches, and cannot be accessed
    directly, but can be iterated on. Each item is represented by a (item,
    notes) tuple, where item is a dict as returned by Zotero, and notes a list
    of the item's child notes (also dicts as returned by Zotero).
    """

    def __init__(self, zotero_credentials, item_types=None):
        """
        Construct the iterable.

        :param zotero.Zotero zotero_credentials: The Zotero instance.

        :param list item_types: List of desired Zotero item types. If None, all
            items will be retrieved.
        """
        self.zotero_credentials = zotero_credentials
        self.item_type_filter = ' || '.join(item_types) if item_types else None
        self.start = current_app.config['KERKO_ZOTERO_START']
        self.zotero_batch = []
        self.iterator = iter(self.zotero_batch)

    def __iter__(self):
        return self

    def __next__(self):
        if current_app.config['KERKO_ZOTERO_END'] > 0 and \
                self.start >= current_app.config['KERKO_ZOTERO_END']:
            raise StopIteration
        while True:
            try:
                return self._next_item()
            except StopIteration:
                self._next_batch()

    @retry_zotero
    def _next_batch(self):
        limit = current_app.config['KERKO_ZOTERO_BATCH_SIZE']
        current_app.logger.info(
            "Requesting up to {limit} Zotero items"
            " starting at position {start}...".format(
                limit=limit, start=self.start
            )
        )
        self.zotero_batch = self.zotero_credentials.items(
            start=self.start,
            limit=limit,
            sort='dateModified',
            direction='asc',
            include='data,coins,bib',
            style=current_app.config['KERKO_CSL_STYLE'],
            itemType=self.item_type_filter
        )
        if not self.zotero_batch:
            raise StopIteration  # Empty batch, nothing more to iterate on.
        self.iterator = iter(self.zotero_batch)

    def _next_item(self):
        zotero_item = next(self.iterator)
        notes = []
        if zotero_item.get('meta', {}).get('numChildren', 0):
            notes = [n for n in ChildNotes(self.zotero_credentials, zotero_item['key'])]
        self.start += 1
        return zotero_item, notes


class ChildNotes:
    """
    Iterable over Zotero notes that are children of an item.

    Notes are loaded on demand, in small batches, and cannot be accessed
    directly, but can be iterated on. Each note is represented by a dict as
    returned by Zotero.
    """

    def __init__(self, zotero_credentials, item_key):
        self.zotero_credentials = zotero_credentials
        self.item_key = item_key
        self.start = 0
        self.zotero_batch = []
        self.iterator = iter(self.zotero_batch)

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            try:
                return self._next_item()
            except StopIteration:
                self._next_batch()

    @retry_zotero
    def _next_batch(self):
        limit = current_app.config['KERKO_ZOTERO_BATCH_SIZE']
        if self.zotero_batch and len(self.zotero_batch) < limit:
            # Previous batch did not reach limit. No need for more batches.
            raise StopIteration
        current_app.logger.info(
            "Requesting up to {limit} child notes for item {item_key}"
            " starting at position {start}...".format(
                limit=limit, item_key=self.item_key, start=self.start
            )
        )
        self.zotero_batch = self.zotero_credentials.children(
            self.item_key,
            start=self.start,
            limit=limit,
            sort='dateModified',
            direction='asc',
            include='data',
            itemType='note',
        )
        if not self.zotero_batch:
            raise StopIteration  # Empty batch, nothing more to iterate.
        self.iterator = iter(self.zotero_batch)

    def _next_item(self):
        zotero_item = next(self.iterator)
        self.start += 1
        return zotero_item



"""
Functions for extracting data from Zotero items.
"""

from abc import ABC, abstractmethod
from collections import defaultdict
import re

from flask import current_app, Markup

from .text import id_normalize, sort_normalize


PHRASE_BREAKER = ' ZZZZZ '  # Some token unlikely to be searched.


class ItemContext:
    """Contains data related to a Zotero item."""

    def __init__(self, item, notes):
        """
        Initialize an item's data.

        :param dict item: An item, as returned by the Zotero API.

        :param list notes: A list of dicts representing the item's child notes.
        """
        self.item_key = item.get('key', '')  # For convenient access.
        self.item = item
        self.notes = notes
        self.data = item.get('data', {})  # For convenient access.


class LibraryContext:
    """Contains data related to a Zotero library."""

    def __init__(self, collections, item_types, item_fields, creator_types):
        self.collections = collections
        self.item_types = item_types
        self.item_fields = item_fields
        self.creator_types = creator_types

    def get_creator_types(self, item_data):
        return self.creator_types.get(item_data.get('itemType', ''), [])


def _parse_zotero_date(text):
    """Parse a fuzzy date into a (year, month, day) tuple of numerical values."""
    year = month = day = 0
    matches = re.match(r'^([0-9]{4})(-([0-9]{2})(-([0-9]{2}))?)?', text)
    if matches:
        if matches.group(1):
            year = int(matches.group(1))
        if matches.group(3):
            month = int(matches.group(3))
        if matches.group(5):
            day = int(matches.group(5))
    return year, month, day


class Extractor(ABC):

    @abstractmethod
    def extract(self, document, spec, item_context, library_context):
        """Extract a value from context and store it in document using spec."""

    def warning(self, message, item_context):
        current_app.logger.warning(
            f"{self.__class__.__name__}: {message} ({item_context.item_key})"
        )


class KeyExtractor(Extractor):  # pylint: disable=abstract-method

    def __init__(self, key):
        self.key = key


class ItemExtractor(KeyExtractor):
    """Extract a value from an item."""

    def extract(self, document, spec, item_context, library_context):
        if self.key in item_context.item:
            document[spec.key] = spec.encode(item_context.item[self.key])


class ItemDataExtractor(KeyExtractor):
    """Extract a value from item data."""

    def extract(self, document, spec, item_context, library_context):
        if self.key in item_context.data:
            document[spec.key] = spec.encode(item_context.data[self.key])


class ItemTypeLabelExtractor(Extractor):
    """Extract the label of the item's type."""

    def extract(self, document, spec, item_context, library_context):
        item_type = item_context.data.get('itemType')
        if item_type and item_type in library_context.item_types:
            document[spec.key] = spec.encode(library_context.item_types[item_type])
        else:
            self.warning("Missing itemType", item_context)


class ItemFieldsExtractor(Extractor):
    """Extract field metadata, serialized as a JSON string."""

    def extract(self, document, spec, item_context, library_context):
        item_type = item_context.data.get('itemType')
        if item_type and item_type in library_context.item_fields:
            fields = library_context.item_fields[item_type]
            # Retain metadata for fields that are actually present in the item.
            item_fields = [f for f in fields if f.get('field') in item_context.data]
            if item_fields:
                document[spec.key] = spec.encode(item_fields)
            else:
                self.warning("Missing item type fields", item_context)
        else:
            self.warning("Missing itemType", item_context)


class CreatorTypesExtractor(Extractor):
    """Extract creator types metadata, serialized as a JSON string."""

    def extract(self, document, spec, item_context, library_context):
        item_type = item_context.data.get('itemType')
        if item_type and item_type in library_context.creator_types:
            library_creator_types = library_context.creator_types[item_type]
            # Retain metadata for creator types that are actually present in the item.
            item_creator_types = []
            for library_creator_type in library_creator_types:
                for item_creator in item_context.data.get('creators', []):
                    creator_type = item_creator.get('creatorType')
                    if creator_type and creator_type == library_creator_type.get('creatorType'):
                        item_creator_types.append(library_creator_type)
                        break
            if item_creator_types:
                document[spec.key] = spec.encode(item_creator_types)
            else:
                self.warning("Missing creator types", item_context)
        else:
            self.warning("Missing itemType", item_context)


class CreatorsExtractor(Extractor):
    """Flatten and extract creator data."""

    def extract(self, document, spec, item_context, library_context):
        creators = []
        if 'creators' in item_context.data:
            for creator in item_context.data['creators']:
                n = creator.get('name', '').strip()
                if n:
                    creators.append(n)
                firstname = creator.get('firstName', '').strip()
                lastname = creator.get('lastName', '').strip()
                if firstname and lastname:
                    # Combine firstname and lastname in different orders to help
                    # phrase searches.
                    creators.append(' '.join([firstname, lastname]))
                    creators.append(', '.join([lastname, firstname]))
                elif firstname:
                    creators.append(firstname)
                elif lastname:
                    creators.append(lastname)
        if creators:
            # Whoosh does not support multivalue fields. And joining multiple
            # values with spaces risks creating phrases matches based on
            # consecutive words. Joining them with an unlikely string
            # (here PHRASE_BREAKER) prevents that.
            # TODO: Avoid this hack by implementing a custom Whoosh field type
            # which could take a list (like the ID field type does) and support
            # phrase search.
            document[spec.key] = spec.encode(PHRASE_BREAKER.join(creators))


class CollectionNamesExtractor(Extractor):
    """Extract item collections for text search."""

    def extract(self, document, spec, item_context, library_context):
        names = set()
        if 'collections' in item_context.data:
            for k in item_context.data['collections']:
                if k in library_context.collections:
                    name = library_context.collections[k].get('data', {}).get('name', '').strip()
                    if name:
                        names.add(name)
        if names:
            document[spec.key] = spec.encode(PHRASE_BREAKER.join(names))


class BaseTagsExtractor(Extractor):

    def __init__(self, whitelist_re='', blacklist_re=''):
        """
        Initialize the extractor.

        :param str whitelist_re: Any tag that does not matches this regular
            expression will be ignored by the extractor. If empty, all tags will
            be accepted unless `blacklist_re` is set and they match it.

        :param str blacklist_re: Any tag that matches this regular expression
            will be ignored by the extractor. If empty, all tags will be
            accepted unless `whitelist_re` is set and they do not match it.
        """
        self.whitelist = re.compile(whitelist_re) if whitelist_re else None
        self.blacklist = re.compile(blacklist_re) if blacklist_re else None

    def extract(self, document, spec, item_context, library_context):
        tags = set()
        if 'tags' in item_context.data:
            for tag_data in item_context.data['tags']:
                tag = tag_data.get('tag', '').strip()
                if tag and \
                        (not self.whitelist or self.whitelist.match(tag)) and \
                        (not self.blacklist or not self.blacklist.match(tag)):
                    tags.add(tag)
        if tags:
            self.to_document(document, spec, tags)

    @abstractmethod
    def to_document(self, document, spec, tags):
        """Assign the extracted value to the document."""


class TagsTextExtractor(BaseTagsExtractor):
    """Extract item tags for text search."""

    def to_document(self, document, spec, tags):
        document[spec.key] = spec.encode(PHRASE_BREAKER.join(tags))


class BaseNotesExtractor(Extractor):

    def __init__(self, whitelist_re='', blacklist_re=''):
        """
        Initialize the extractor.

        :param str whitelist_re: Any note which does not have a tag that matches
            this regular expression will be ignored by the extractor. If empty,
            all notes will be accepted unless `blacklist_re` is set and causes
            some to be rejected.

        :param str blacklist_re: Any note that have a tag that matches this
            regular expression will be ignored by the extractor. If empty, all
            notes will be accepted unless `whitelist_re` is set and causes some
            to be rejected.
        """
        self.whitelist = re.compile(whitelist_re) if whitelist_re else None
        self.blacklist = re.compile(blacklist_re) if blacklist_re else None

    def extract(self, document, spec, item_context, library_context):
        accepted_notes = []
        for note in item_context.notes:
            whitelisted = self.whitelist is None
            blacklisted = False
            if self.whitelist or self.blacklist:
                for tag_data in note.get('data', {}).get('tags', []):
                    tag = tag_data.get('tag', '').strip()
                    if self.whitelist and self.whitelist.match(tag):
                        whitelisted = True
                    if self.blacklist and self.blacklist.match(tag):
                        blacklisted = True
            if whitelisted and not blacklisted:
                accepted_notes.append(note.get('data', {}).get('note', ''))
        if accepted_notes:
            self.to_document(document, spec, accepted_notes)

    @abstractmethod
    def to_document(self, document, spec, notes):
        """Assign the extracted notes to the document."""


class NotesTextExtractor(BaseNotesExtractor):
    """Extract notes for text search."""

    def to_document(self, document, spec, notes):
        document[spec.key] = spec.encode(
            PHRASE_BREAKER.join([Markup(n).striptags() for n in notes])
        )


class RawNotesExtractor(BaseNotesExtractor):
    """Extract raw notes for storage."""

    def to_document(self, document, spec, notes):
        document[spec.key] = spec.encode(notes)


class RawDataExtractor(Extractor):

    def extract(self, document, spec, item_context, library_context):
        document[spec.key] = spec.encode(item_context.data)


def _expand_paths(path):
    """
    Extract the paths of each of the components of the specified path.

    If the given path is ['a', 'b', 'c'], the returned list of paths is:
    [['a'], ['a', 'b'], ['a', 'b', 'c']]
    """
    return [path[0:i + 1] for i in range(len(path))]


class CollectionFacetTreeExtractor(Extractor):
    """Index the Zotero item's collections needed for the specified facet."""

    def extract(self, document, spec, item_context, library_context):
        # Sets prevent duplication when multiple collections share common ancestors.
        encoded_ancestors = defaultdict(set)

        for collection_key in item_context.data.get('collections', []):
            if collection_key not in library_context.collections:
                continue  # Skip unknown collection.
            ancestors = library_context.collections.ancestors(collection_key)
            if len(ancestors) <= 1 or ancestors[0] != spec.collection_key:
                continue  # Skip collection, unrelated to this facet.

            ancestors = ancestors[1:]  # Facet values come from subcollections.
            for path in _expand_paths(ancestors):
                label = library_context.collections.get(
                    path[-1], {}
                ).get('data', {}).get('name', '').strip()
                encoded_ancestors[spec.key].add(spec.encode((path, label)))

        for key, ancestors in encoded_ancestors.items():
            document[key] = list(ancestors)


class TagsFacetExtractor(BaseTagsExtractor):
    """Index the Zotero item's tags for faceting."""

    def to_document(self, document, spec, tags):
        document[spec.key] = [spec.encode(tag) for tag in tags]


class ItemTypeFacetExtractor(Extractor):
    """Index the Zotero item's type for faceting."""

    def extract(self, document, spec, item_context, library_context):
        item_type = item_context.data.get('itemType')
        if item_type:
            document[spec.key] = spec.encode(
                (item_type, library_context.item_types.get(item_type, item_type))
            )
        else:
            self.warning("Missing itemType", item_context)


class YearFacetExtractor(Extractor):
    """Index the Zotero item's publication date for faceting by year."""

    def extract(self, document, spec, item_context, library_context):
        parsed_date = item_context.item.get('meta', {}).get('parsedDate', '')
        if parsed_date:
            year, _month, _day = _parse_zotero_date(parsed_date)
            decade = int(int(year) / 10) * 10
            century = int(int(year) / 100) * 100
            encoded_paths = [
                spec.encode(path) for path in _expand_paths(
                    [str(century), str(decade), str(year)]
                )
            ]
            document[spec.key] = encoded_paths


class ItemDataLinkFacetExtractor(ItemDataExtractor):

    def extract(self, document, spec, item_context, library_context):
        document[spec.key] = spec.encode(item_context.data.get(self.key, '').strip() != '')


def _prepare_sort_text(text):
    """
    Normalize the given text for a sort field.

    Sort fields are bytearrays in the schema, so the text goes through
    str.encode().

    :param str text: The Unicode string to normalize.

    :return bytearray: The normalized text.
    """
    return sort_normalize(Markup(text).striptags()).encode()


class SortItemDataExtractor(ItemDataExtractor):

    def extract(self, document, spec, item_context, library_context):
        document[spec.key] = spec.encode(
            _prepare_sort_text(item_context.data.get(self.key, ''))
        )


class SortCreatorExtractor(Extractor):

    def extract(self, document, spec, item_context, library_context):
        creators = []

        def append_creator(creator):
            creator_parts = [
                _prepare_sort_text(creator.get('lastName', '')),
                _prepare_sort_text(creator.get('firstName', '')),
                _prepare_sort_text(creator.get('name', ''))]
            creators.append(b' zzz '.join([p for p in creator_parts if p]))

        # We treat creator types like an ordered list, where the first creator
        # type is for primary creators. Depending on the citation style, lesser
        # creator types may not appear in citations. Therefore, we try to sort
        # only by primary creators in order to avoid sorting with data that may
        # be invisible to the user. Only when an item has no primary creator do
        # we fallback to lesser creators.
        for creator_type in library_context.get_creator_types(item_context.data):
            for creator in item_context.data.get('creators', []):
                if creator.get('creatorType', '') == creator_type.get('creatorType'):
                    append_creator(creator)
            if creators:
                break  # No need to include lesser creator types.
        document[spec.key] = spec.encode(b' zzzzzz '.join(creators))


class SortDateExtractor(Extractor):

    def extract(self, document, spec, item_context, library_context):
        parsed_date = item_context.item.get('meta', {}).get('parsedDate', '')
        year, month, day = _parse_zotero_date(parsed_date)
        document[spec.key] = spec.encode(
            int('{:04d}{:02d}{:02d}'.format(year, month, day))
        )



"""
Kerko: A Flask blueprint that provides faceted search for bibliographies based
on Zotero.
"""
# pylint: disable=invalid-name

import pathlib

from environs import Env
from flask import Blueprint
from flask_babelex import Domain

__version__ = '0.3alpha0'

babel_domain = Domain(
    pathlib.Path(__file__).parent / 'translations',
    domain='kerko'
)
env = Env()


def init_default_config(state):
    """
    Initialize default KERKO settings in the app's config object.

    The following settings must be set in the app's config and have no defaults:

    - `KERKO_ZOTERO_LIBRARY_ID`
    - `KERKO_ZOTERO_LIBRARY_TYPE`
    - `KERKO_ZOTERO_API_KEY`

    The following settings are used with Bootstrap-Flask to load CDN-based
    resources. Alternatively, one may set `BOOTSTRAP_SERVE_LOCAL=True` to use
    the static resources included with the Bootstrap-Flask extension (and ignore
    these settings):

    - `KERKO_BOOTSTRAP_VERSION`
    - `KERKO_JQUERY_VERSION`
    - `KERKO_POPPER_VERSION`
    - `KERKO_WITH_JQUERY`
    - `KERKO_WITH_POPPER`
    """
    app_dir = pathlib.Path(env.str('FLASK_APP')).parent.absolute()

    state.app.config.setdefault('KERKO_TITLE', 'Kerko')
    state.app.config.setdefault('KERKO_DATA_DIR', app_dir / 'data' / 'kerko')
    state.app.config.setdefault('KERKO_WHOOSH_LANGUAGE', 'en')
    state.app.config.setdefault('KERKO_ZOTERO_LOCALE', 'en-US')
    state.app.config.setdefault('KERKO_ZOTERO_START', 0)
    state.app.config.setdefault('KERKO_ZOTERO_END', 0)
    state.app.config.setdefault('KERKO_ZOTERO_MAX_ATTEMPTS', 10)
    state.app.config.setdefault('KERKO_ZOTERO_WAIT', 120)  # In seconds.
    state.app.config.setdefault('KERKO_ZOTERO_BATCH_SIZE', 100)
    state.app.config.setdefault('KERKO_PAGE_LEN', 20)
    state.app.config.setdefault('KERKO_PAGER_LINKS', 4)
    state.app.config.setdefault('KERKO_CSL_STYLE', 'apa')
    state.app.config.setdefault('KERKO_RESULTS_ABSTRACT', False)
    state.app.config.setdefault('KERKO_FACET_COLLAPSING', False)
    state.app.config.setdefault('KERKO_PRINT_ITEM_LINK', False)
    state.app.config.setdefault('KERKO_PRINT_CITATIONS_LINK', False)
    state.app.config.setdefault('KERKO_PRINT_CITATIONS_MAX_COUNT', 0)
    state.app.config.setdefault('KERKO_USE_TRANSLATIONS', True)

    state.app.config.setdefault('KERKO_BOOTSTRAP_VERSION', '4.3.1')
    state.app.config.setdefault('KERKO_JQUERY_VERSION', '3.3.1')
    state.app.config.setdefault('KERKO_POPPER_VERSION', '1.14.7')
    state.app.config.setdefault('KERKO_WITH_JQUERY', True)
    state.app.config.setdefault('KERKO_WITH_POPPER', True)


blueprint = Blueprint(
    'kerko', __name__, static_folder='static', template_folder='templates'
)
blueprint.record_once(init_default_config)


from . import views  # pylint: disable=wrong-import-position



from flask_babelex import lazy_gettext as _
import whoosh
from whoosh.analysis import CharsetFilter, LowercaseFilter, StemFilter
from whoosh.analysis.tokenizers import RegexTokenizer
from whoosh.fields import columns, Schema, BOOLEAN, COLUMN, ID, NUMERIC, TEXT, STORED
from whoosh.query import Prefix, Term
from whoosh.support.charset import accent_map

from . import codecs, extractors
from .specs import ScopeSpec, FieldSpec, FlatFacetSpec, TreeFacetSpec, SortSpec


class Composer:
    """
    A factory for the setting up the search elements.

    This class acts as a registry for configuring the search elements such as
    fields and facets, from which both the schema and the search interface can
    be built.

    The schema is the representation of documents and their fields in the search
    index. It is meant to be fully constructed at app configuration time and not
    changed afterwards. If schema elements need to be modified or removed, the
    application should be stopped and the index cleaned and rebuilt.
    """

    def __init__(
            self,
            whoosh_language='en',
            exclude_default_scopes=None,
            exclude_default_fields=None,
            exclude_default_facets=None,
            exclude_default_sorts=None,
            default_item_whitelist_tag_re='',
            default_item_blacklist_tag_re=r'^_',
            default_note_whitelist_tag_re='',
            default_note_blacklist_tag_re=r'^_',
    ):  # pylint: disable=too-many-arguments
        """
        Initialize the object.

        Based on the provided arguments, default fields and facets may be
        initialized. Those may be modified after initialization; it is safe to
        change attributes that impact the search interface and not the schema,
        but those changes that should impact the schema require the field to be
        removed first, then re-added with the changes.

        :param list exclude_default_scopes: List of scopes (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no scope will be added by default. Caution: default
            fields are expecting those scopes to exist. You'll have to add them
            manually or alter the fields.

        :param list exclude_default_fields: List of fields (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no field will be created by default. Caution: some
            default fields are required by Kerko. If those are excluded, you'll
            have to add them manually.

        :param list exclude_default_facets: List of facets (identified by key)
            to exclude from those created by default. If that list contains the
            value '*', no facet will be created by default.

        :param list exclude_default_sorts: List of sorts (identified by key) to
            exclude from those created by default. If that list contains the
            value '*', no sort will be created by default and you'll have to
            manually add at least one sort.
        """
        if not whoosh.lang.has_stemmer(whoosh_language):
            whoosh_language = 'en'

        # Replace the standard analyzer with one that has no stop words (helping
        # people who do phrase searches without specifying actual phrase queries).
        self.text_chain = RegexTokenizer() | StemFilter(lang=whoosh_language) | CharsetFilter(accent_map) | LowercaseFilter()

        # Same for names, but without stemming.
        self.name_chain = RegexTokenizer() | CharsetFilter(accent_map) | LowercaseFilter()

        # Common schema field types.
        self.id_field_type = ID(field_boost=50.0)
        self.secondary_id_field_type = ID(field_boost=10.0)
        self.title_field_type = TEXT(analyzer=self.text_chain, field_boost=5.0)
        self.secondary_title_field_type = TEXT(analyzer=self.text_chain, field_boost=2.0)
        self.creator_name_field_type = TEXT(analyzer=self.name_chain, field_boost=2.0)
        self.name_field_type = TEXT(analyzer=self.name_chain)
        self.text_field_type = TEXT(analyzer=self.text_chain)

        self.schema = Schema()
        self.scopes = {}
        self.fields = {}
        self.facets = {}
        self.sorts = {}
        self.exclude_default_scopes = exclude_default_scopes or []
        self.exclude_default_fields = exclude_default_fields or []
        self.exclude_default_facets = exclude_default_facets or []
        self.exclude_default_sorts = exclude_default_sorts or []
        self.default_item_whitelist_tag_re = default_item_whitelist_tag_re
        self.default_item_blacklist_tag_re = default_item_blacklist_tag_re
        self.default_note_whitelist_tag_re = default_note_whitelist_tag_re
        self.default_note_blacklist_tag_re = default_note_blacklist_tag_re
        self.init_default_scopes()
        self.init_default_fields()
        self.init_default_facets()
        self.init_default_sorts()

    def init_default_scopes(self):
        if '*' in self.exclude_default_scopes:
            return

        if 'all' not in self.exclude_default_scopes:
            self.add_scope(
                ScopeSpec(
                    key='all',
                    selector_label=_("In any field"),
                    breadbox_label=_("Any field"),
                    weight=0,
                )
            )
        if 'creator' not in self.exclude_default_scopes:
            self.add_scope(
                ScopeSpec(
                    key='creator',
                    selector_label=_("In authors/contributors"),
                    breadbox_label=_("Author/contributor"),
                    weight=100,
                )
            )
        if 'title' not in self.exclude_default_scopes:
            self.add_scope(
                ScopeSpec(
                    key='title',
                    selector_label=_("In titles"),
                    breadbox_label=_("Title"),
                    weight=200,
                )
            )

    def init_default_fields(self):
        if '*' in self.exclude_default_fields:
            return

        # Document id for the search index, matching the Zotero item id.
        if 'id' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='id',
                    field_type=ID(unique=True, stored=True),
                    extractor=extractors.ItemExtractor('key')
                )
            )
        # Label of this item's type.
        if 'item_type' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='item_type',
                    field_type=TEXT(analyzer=self.text_chain, stored=True),
                    scopes=['all'],
                    extractor=extractors.ItemTypeLabelExtractor(),
                )
            )

        # All Zotero item fields that we want to make available to keyword
        # search are specified below, each with an appropriate analyzer chain.
        # The full list of item fields can be obtained through the Zotero API.
        # Keys derive from the corresponding field names in Zotero, but are
        # prefixed with 'z_' to prevent name clashes with Kerko's own fields.

        #
        # Identifier fields, for keyword search. Highly boosted for matches to
        # rank at the top of results.
        #

        if 'z_DOI' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_DOI',
                    field_type=self.id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('DOI')
                )
            )
        if 'z_ISBN' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_ISBN',
                    field_type=self.id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('ISBN')
                )
            )
        if 'z_ISSN' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_ISSN',
                    field_type=self.id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('ISSN')
                )
            )

        #
        # Secondary identifiers, for keyword search. Moderately boosted.
        #

        if 'z_applicationNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_applicationNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('applicationNumber')
                )
            )
        if 'z_billNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_billNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('billNumber')
                )
            )
        if 'z_callNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_callNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('callNumber')
                )
            )
        if 'z_codeNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_codeNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('codeNumber')
                )
            )
        if 'z_docketNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_docketNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('docketNumber')
                )
            )
        if 'z_documentNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_documentNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('documentNumber')
                )
            )
        if 'z_patentNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_patentNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('patentNumber')
                )
            )
        if 'z_priorityNumbers' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_priorityNumbers',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('priorityNumbers')
                )
            )
        if 'z_publicLawNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_publicLawNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('publicLawNumber')
                )
            )
        if 'z_reportNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_reportNumber',
                    field_type=self.secondary_id_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('reportNumber')
                )
            )

        #
        # Title fields, for keyword search.
        #

        if 'z_nameOfAct' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_nameOfAct',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor('nameOfAct')
                )
            )
        if 'z_shortTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_shortTitle',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor('shortTitle')
                )
            )
        if 'z_subject' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_subject',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor('subject')
                )
            )
        if 'z_title' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_title',
                    field_type=self.title_field_type,
                    scopes=['all', 'title'],
                    extractor=extractors.ItemDataExtractor('title')
                )
            )

        #
        # Container titles, for keyword search.
        #

        if 'z_blogTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_blogTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('blogTitle')
                )
            )
        if 'z_bookTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_bookTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('bookTitle')
                )
            )
        if 'z_code' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_code',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('code')
                )
            )
        if 'z_conferenceName' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_conferenceName',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('conferenceName')
                )
            )
        if 'z_dictionaryTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_dictionaryTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('dictionaryTitle')
                )
            )
        if 'z_encyclopediaTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_encyclopediaTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('encyclopediaTitle')
                )
            )
        if 'z_forumTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_forumTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('forumTitle')
                )
            )
        if 'z_meetingName' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_meetingName',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('meetingName')
                )
            )
        if 'z_proceedingsTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_proceedingsTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('proceedingsTitle')
                )
            )
        if 'z_programTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_programTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('programTitle')
                )
            )
        if 'z_publicationTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_publicationTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('publicationTitle')
                )
            )
        if 'z_section' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_section',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('section')
                )
            )
        if 'z_series' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_series',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('series')
                )
            )
        if 'z_seriesTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_seriesTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('seriesTitle')
                )
            )
        if 'z_session' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_session',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('session')
                )
            )
        if 'z_websiteTitle' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_websiteTitle',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('websiteTitle')
                )
            )

        #
        # Name fields, for keyword search. Exempt from stemming.
        #

        if 'z_archive' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_archive',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('archive')
                )
            )
        if 'z_archiveLocation' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_archiveLocation',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('archiveLocation')
                )
            )
        if 'z_assignee' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_assignee',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('assignee')
                )
            )
        if 'z_audioFileType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_audioFileType',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('audioFileType')
                )
            )
        if 'z_audioRecordingFormat' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_audioRecordingFormat',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('audioRecordingFormat')
                )
            )
        if 'z_caseName' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_caseName',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('caseName')
                )
            )
        if 'z_committee' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_committee',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('committee')
                )
            )
        if 'z_company' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_company',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('company')
                )
            )
        if 'z_country' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_country',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('country')
                )
            )
        if 'z_court' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_court',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('court')
                )
            )
        if 'z_distributor' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_distributor',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('distributor')
                )
            )
        if 'z_institution' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_institution',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('institution')
                )
            )
        if 'z_issuingAuthority' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_issuingAuthority',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('issuingAuthority')
                )
            )
        if 'z_journalAbbreviation' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_journalAbbreviation',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('journalAbbreviation')
                )
            )
        if 'z_label' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_label',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('label')
                )
            )
        if 'z_legislativeBody' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_legislativeBody',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('legislativeBody')
                )
            )
        if 'z_libraryCatalog' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_libraryCatalog',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('libraryCatalog')
                )
            )
        if 'z_network' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_network',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('network')
                )
            )
        if 'z_place' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_place',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('place')
                )
            )
        if 'z_programmingLanguage' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_programmingLanguage',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('programmingLanguage')
                )
            )
        if 'z_publisher' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_publisher',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('publisher')
                )
            )
        if 'z_reporter' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_reporter',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('reporter')
                )
            )
        if 'z_studio' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_studio',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('studio')
                )
            )
        if 'z_system' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_system',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('system')
                )
            )
        if 'z_university' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_university',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('university')
                )
            )
        if 'z_videoRecordingFormat' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_videoRecordingFormat',
                    field_type=self.name_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('videoRecordingFormat')
                )
            )

        #
        # Date fields, for keyword search.
        #

        if 'z_dateDecided' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_dateDecided',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('dateDecided')
                )
            )
        if 'z_dateEnacted' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_dateEnacted',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('dateEnacted')
                )
            )
        if 'z_date' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_date',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('date')
                )
            )
        if 'z_filingDate' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_filingDate',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('filingDate')
                )
            )
        if 'z_issueDate' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_issueDate',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('issueDate')
                )
            )

        #
        # Text fields, for keyword search.
        #

        if 'z_abstractNote' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_abstractNote',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('abstractNote')
                )
            )
        if 'z_artworkMedium' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_artworkMedium',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('artworkMedium')
                )
            )
        if 'z_artworkSize' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_artworkSize',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('artworkSize')
                )
            )
        if 'z_codeVolume' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_codeVolume',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('codeVolume')
                )
            )
        if 'z_edition' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_edition',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('edition')
                )
            )
        if 'z_episodeNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_episodeNumber',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('episodeNumber')
                )
            )
        if 'z_genre' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_genre',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('genre')
                )
            )
        if 'z_history' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_history',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('history')
                )
            )
        if 'z_interviewMedium' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_interviewMedium',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('interviewMedium')
                )
            )
        if 'z_issue' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_issue',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('issue')
                )
            )
        if 'z_language' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_language',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('language')
                )
            )
        if 'z_legalStatus' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_legalStatus',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('legalStatus')
                )
            )
        if 'z_letterType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_letterType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('letterType')
                )
            )
        if 'z_manuscriptType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_manuscriptType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('manuscriptType')
                )
            )
        if 'z_mapType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_mapType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('mapType')
                )
            )
        if 'z_postType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_postType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('postType')
                )
            )
        if 'z_presentationType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_presentationType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('presentationType')
                )
            )
        if 'z_references' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_references',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('references')
                )
            )
        if 'z_reporterVolume' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_reporterVolume',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('reporterVolume')
                )
            )
        if 'z_reportType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_reportType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('reportType')
                )
            )
        if 'z_rights' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_rights',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('rights')
                )
            )
        if 'z_seriesNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_seriesNumber',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('seriesNumber')
                )
            )
        if 'z_seriesText' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_seriesText',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('seriesText')
                )
            )
        if 'z_thesisType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_thesisType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('thesisType')
                )
            )
        if 'z_versionNumber' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_versionNumber',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('versionNumber')
                )
            )
        if 'z_volume' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_volume',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('volume')
                )
            )
        if 'z_websiteType' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='z_websiteType',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.ItemDataExtractor('websiteType')
                )
            )

        # Known Zotero item fields that we are deliberately ignoring for search:
        # - accessDate
        # - codePages
        # - extra
        # - firstPage
        # - numberOfVolumes
        # - numPages
        # - pages
        # - runningTime
        # - scale
        # - url

        #
        # Other metadata associated to Zotero items, for keyword search.
        #

        # Creators, exempt from stemming.
        if 'text_creator' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='text_creator',
                    field_type=self.creator_name_field_type,
                    scopes=['all', 'creator'],
                    extractor=extractors.CreatorsExtractor()
                )
            )
        if 'text_collections' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='text_collections',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.CollectionNamesExtractor()
                )
            )
        if 'text_tags' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='text_tags',
                    field_type=self.secondary_title_field_type,
                    scopes=['all'],
                    extractor=extractors.TagsTextExtractor(
                        whitelist_re=self.default_item_whitelist_tag_re,
                        blacklist_re=self.default_item_blacklist_tag_re
                    )
                )
            )
        if 'text_notes' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='text_notes',
                    field_type=self.text_field_type,
                    scopes=['all'],
                    extractor=extractors.NotesTextExtractor(
                        whitelist_re=self.default_note_whitelist_tag_re,
                        blacklist_re=self.default_note_blacklist_tag_re
                    )
                )
            )

        #
        # Stored fields, not available for keyword search (hence the lack of a
        # 'scopes' parameter).
        #

        # Formatted citation.
        if 'bib' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='bib',
                    field_type=STORED,
                    extractor=extractors.ItemExtractor('bib')
                )
            )
        # OpenURL Coins.
        if 'coins' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='coins',
                    field_type=STORED,
                    extractor=extractors.ItemExtractor('coins')
                )
            )
        # Raw item data.
        if 'data' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='data',
                    field_type=STORED,
                    extractor=extractors.RawDataExtractor(),
                    codec=codecs.JSONFieldCodec()
                )
            )
        # Child notes of the item.
        if 'notes' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='notes',
                    field_type=STORED,
                    extractor=extractors.RawNotesExtractor(
                        whitelist_re=self.default_note_whitelist_tag_re,
                        blacklist_re=self.default_note_blacklist_tag_re
                    )
                )
            )
        # Fields and labels for this item type, for convenient access.
        if 'item_fields' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='item_fields',
                    field_type=STORED,
                    extractor=extractors.ItemFieldsExtractor(),
                    codec=codecs.JSONFieldCodec()
                )
            )
        # Creator types for this item type, for convenient access.
        if 'creator_types' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='creator_types',
                    field_type=STORED,
                    extractor=extractors.CreatorTypesExtractor(),
                    codec=codecs.JSONFieldCodec()
                )
            )

        #
        # Fields for sorting.
        #

        if 'sort_title' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='sort_title',
                    field_type=COLUMN(columns.VarBytesColumn()),
                    extractor=extractors.SortItemDataExtractor('title'),
                )
            )
        if 'sort_creator' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='sort_creator',
                    field_type=COLUMN(columns.VarBytesColumn()),
                    extractor=extractors.SortCreatorExtractor(),
                )
            )
        if 'sort_date' not in self.exclude_default_fields:
            self.add_field(
                FieldSpec(
                    key='sort_date',
                    field_type=NUMERIC,
                    extractor=extractors.SortDateExtractor(),
                )
            )

    def init_default_facets(self):
        if '*' in self.exclude_default_facets:
            return

        if 'facet_tag' not in self.exclude_default_facets:
            self.add_facet(
                FlatFacetSpec(
                    key='facet_tag',
                    title=_('Topic'),
                    filter_key='topic',
                    weight=100,
                    field_type=ID(stored=True),
                    extractor=extractors.TagsFacetExtractor(
                        whitelist_re=self.default_item_whitelist_tag_re,
                        blacklist_re=self.default_item_blacklist_tag_re
                    ),
                    codec=codecs.BaseFacetCodec(),
                    missing_label=None,
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=True,
                    allow_overlap=True,
                    query_class=Term
                )
            )
        if 'facet_item_type' not in self.exclude_default_facets:
            self.add_facet(
                FlatFacetSpec(
                    key='facet_item_type',
                    title=_('Resource type'),
                    filter_key='type',
                    weight=200,
                    field_type=ID(stored=True),
                    extractor=extractors.ItemTypeFacetExtractor(),
                    codec=codecs.ItemTypeFacetCodec(),
                    missing_label=None,
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=False,
                    allow_overlap=False,
                    query_class=Prefix
                )
            )
        if 'facet_year' not in self.exclude_default_facets:
            self.add_facet(
                TreeFacetSpec(
                    key='facet_year',
                    title=_('Publication year'),
                    filter_key='year',
                    weight=300,
                    field_type=ID(stored=True),
                    extractor=extractors.YearFacetExtractor(),
                    codec=codecs.YearTreeFacetCodec(),
                    missing_label=_('Unknown'),
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=False,
                    allow_overlap=True,
                    query_class=Prefix
                )
            )
        if 'facet_link' not in self.exclude_default_facets:
            self.add_facet(
                FlatFacetSpec(
                    key='facet_link',
                    title=_('Online resource'),
                    filter_key='link',
                    weight=400,
                    field_type=BOOLEAN(stored=True),
                    extractor=extractors.ItemDataLinkFacetExtractor('url'),
                    codec=codecs.BooleanFacetCodec(),
                    missing_label=None,
                    sort_key=['label'],
                    sort_reverse=False,
                    item_view=False,
                    allow_overlap=False,
                    query_class=Term
                )
            )

    def init_default_sorts(self):
        """
        Initialize a set of default `SortSpec` instances.

        These rely on `FieldSpec` instances, which must have been added first.
        """
        if '*' in self.exclude_default_sorts:
            return

        if 'score' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='score',
                    label=_('Relevance'),
                    weight=0,
                    fields=None,
                    # Sort by score is only possible on keyword search.
                    is_allowed=lambda criteria: criteria.has_keyword_search()
                )
            )
        if 'date_desc' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='date_desc',
                    label=_('Newest first'),
                    weight=10,
                    fields=[
                        self.fields['sort_date'],
                        self.fields['sort_creator'],
                        self.fields['sort_title']
                    ],
                    reverse=True
                )
            )
        if 'date_asc' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='date_asc',
                    label=_('Oldest first'),
                    weight=11,
                    fields=[
                        self.fields['sort_date'],
                        self.fields['sort_creator'],
                        self.fields['sort_title']
                    ]
                )
            )
        if 'author_asc' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='author_asc',
                    label=_('Author A-Z'),
                    weight=20,
                    fields=[
                        self.fields['sort_creator'],
                        self.fields['sort_title'],
                        self.fields['sort_date'],
                    ]
                )
            )
        if 'author_desc' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='author_desc',
                    label=_('Author Z-A'),
                    weight=21,
                    fields=[
                        self.fields['sort_creator'],
                        self.fields['sort_title'],
                        self.fields['sort_date'],
                    ],
                    reverse=True
                )
            )
        if 'title_asc' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='title_asc',
                    label=_('Title A-Z'),
                    weight=30,
                    fields=[
                        self.fields['sort_title'],
                        self.fields['sort_creator'],
                        self.fields['sort_date']
                    ]
                )
            )
        if 'title_desc' not in self.exclude_default_sorts:
            self.add_sort(
                SortSpec(
                    key='title_desc',
                    label=_('Title Z-A'),
                    weight=31,
                    fields=[
                        self.fields['sort_title'],
                        self.fields['sort_creator'],
                        self.fields['sort_date']
                    ],
                    reverse=True
                )
            )

    def add_scope(self, scope):
        self.scopes[scope.key] = scope

    def remove_scope(self, key):
        del self.scopes[key]

    def get_ordered_scopes(self):
        """Return a list of all scope specifications, sorted by weight."""
        return sorted(self.scopes.values(), key=lambda spec: spec.weight)

    def add_field(self, field):
        self.fields[field.key] = field
        self.schema.add(field.key, field.field_type)

    def remove_field(self, key):
        self.schema.remove(key)
        del self.fields[key]

    def add_facet(self, facet):
        self.facets[facet.key] = facet
        self.schema.add(facet.key, facet.field_type)

    def remove_facet(self, key):
        self.schema.remove(key)
        del self.facets[key]

    def get_ordered_facets(self):
        """Return a list of all facet specifications, sorted by weight."""
        return sorted(self.facets.values(), key=lambda spec: spec.weight)

    def get_facet_by_filter_key(self, filter_key, default=None):
        for spec in self.facets.values():
            if spec.filter_key == filter_key:
                return spec
        return default

    def add_sort(self, sort):
        self.sorts[sort.key] = sort

    def remove_sort(self, key):
        del self.sorts[key]

    def get_ordered_sorts(self):
        """Return a list of all sort specifications, sorted by weight."""
        return sorted(self.sorts.values(), key=lambda spec: spec.weight)



from babel.numbers import format_number
from flask import current_app
from flask_babelex import get_locale


def build_pager(criteria, page_count, page_len):
    """Build pager items for use in a search view."""
    if page_count <= 1:
        return None

    pager = {}
    page_num = min(criteria.page_num, page_count)
    first = max(1, int(page_num - current_app.config['KERKO_PAGER_LINKS'] / 2))
    last = min(page_count + 1, first + current_app.config['KERKO_PAGER_LINKS'] + 1)
    first = max(1, last - current_app.config['KERKO_PAGER_LINKS'] - 1)

    if page_num > 1:
        p = 1
        pager['first'] = {
            'page_num': p,
            'page_num_formatted': format_number(p, locale=get_locale()),
            'url': criteria.build_url(page_num=p, page_len=page_len),
        }
        p = max(1, page_num - 1)
        pager['previous'] = {
            'page_num': p,
            'page_num_formatted': format_number(p, locale=get_locale()),
            'url': criteria.build_url(page_num=p, page_len=page_len),
        }

    pager['before'] = []
    for p in range(first, page_num):
        pager['before'].append({
            'page_num': p,
            'page_num_formatted': format_number(p, locale=get_locale()),
            'url': criteria.build_url(page_num=p, page_len=page_len),
        })

    pager['current'] = {
        'page_num': page_num,
        'page_num_formatted': format_number(page_num, locale=get_locale()),
        'url': criteria.build_url(page_num=page_num, page_len=page_len),
    }

    pager['after'] = []
    for p in range(page_num + 1, last):
        pager['after'].append({
            'page_num': p,
            'page_num_formatted': format_number(p, locale=get_locale()),
            'url': criteria.build_url(page_num=p, page_len=page_len),
        })

    if page_num < page_count:
        p = min(page_count, page_num + 1)
        pager['next'] = {
            'page_num': p,
            'page_num_formatted': format_number(p, locale=get_locale()),
            'url': criteria.build_url(page_num=p, page_len=page_len),
        }
        p = page_count
        pager['last'] = {
            'page_num': p,
            'page_num_formatted': format_number(p, locale=get_locale()),
            'url': criteria.build_url(page_num=p, page_len=page_len),
        }

    return pager



#!/usr/bin/env python

import io
import re
from setuptools import setup, find_packages
from setuptools.command.install import install as BaseInstallCommand

with io.open('README.md', 'rt', encoding='utf8') as f:
    readme = f.read()

with io.open('kerko/__init__.py', 'rt', encoding='utf8') as f:
    version = re.search(r"__version__ = '(.*?)'", f.read()).group(1)


class InstallCommand(BaseInstallCommand):

    def run(self):
        self.run_command('compile_catalog')  # Compile gettext MO files.
        super().run()


setup(
    name="Kerko",
    version=version,
    url="https://github.com/whiskyechobravo/kerko",
    project_urls={
        "Documentation": "https://github.com/whiskyechobravo/kerko",
        "Code": "https://github.com/whiskyechobravo/kerko",
        "Issue tracker": "https://github.com/whiskyechobravo/kerko/issues",
    },
    author="David Lesieur",
    author_email="kerko@whiskyechobravo.com",
    description="A Flask blueprint that provides a faceted search interface "
                "for bibliographies based on Zotero.",
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=[
        "Babel>=2.6.0",
        "Bootstrap-Flask>=1.0.10",
        "environs>=4.0.0",
        "Flask>=1.0.2",
        "Flask-BabelEx>=0.9.3",
        "Flask-WTF>=0.14.2",
        "Jinja2>=2.10.1",
        "Pyzotero>=1.4.1",
        "Werkzeug>=0.15",
        "Whoosh>=2.7.4",
        "wrapt>=1.10.0",
        "WTForms>=2.2",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Database :: Front-Ends",
        "Topic :: Education",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points={
        'flask.commands': ['kerko=kerko.cli:cli'],
    },
    message_extractors={
        'kerko': [
            ('**.py', 'python', None),
            ('**.jinja2', 'jinja2', {
                'extensions': 'jinja2.ext.autoescape, jinja2.ext.with_, jinja2.ext.do, jinja2.ext.i18n',
            })
        ]
    },
    cmdclass={
        'install': InstallCommand,
    }
)



import time

from babel.numbers import format_number
from flask import abort, current_app, redirect, request, render_template, url_for
from flask_babelex import get_locale, gettext, ngettext

from . import babel_domain, blueprint
from .breadbox import build_breadbox
from .criteria import Criteria
from .forms import SearchForm
from .query import run_query, run_query_unique, build_creators_display, build_fake_facet_results
from .pager import build_pager
from .sorter import build_sorter


@blueprint.route('/', methods=['GET', 'POST'])
def search():
    start_time = time.process_time()

    if current_app.config['KERKO_USE_TRANSLATIONS']:
        babel_domain.as_default()
    criteria = Criteria(request)

    form = SearchForm()
    if form.validate_on_submit():
        url = criteria.build_add_keywords_url(
            scope=form.scope.data,
            value=form.keywords.data)
        return redirect(url, 302)

    if criteria.page_len != 1:  # Note: page_len can be None (for all results).
        return_fields = ['id', 'bib', 'coins']
        if current_app.config['KERKO_RESULTS_ABSTRACT']:
            return_fields.append('data')
    else:
        return_fields = None  # All fields

    search_results, facet_results, total_count, page_count = run_query(
        criteria, return_fields
    )

    breadbox = build_breadbox(criteria, facet_results)
    context = {
        'facet_results': facet_results,
        'breadbox': breadbox,
        'active_facets': breadbox['filters'].keys() if 'filters' in breadbox else [],
        'pager': build_pager(criteria, page_count, criteria.page_len),
        'sorter': build_sorter(criteria),
        'total_count': total_count,
        'total_count_formatted': format_number(total_count, locale=get_locale()),
        'page_count': page_count,
        'page_count_formatted': format_number(page_count, locale=get_locale()),
        'page_len': criteria.page_len,
        'is_searching': criteria.has_keyword_search() or criteria.has_filter_search(),
        'locale': get_locale(),
    }

    if criteria.page_len == 1 and total_count != 0:
        list_page_num = int((criteria.page_num - 1) / current_app.config['KERKO_PAGE_LEN'] + 1)
        build_creators_display(search_results[0])
        build_fake_facet_results(search_results[0])
        return render_template(
            'kerko/search-item.html.jinja2',
            title=search_results[0].get('data', {}).get('title', ''),
            item=search_results[0],
            item_url=url_for(
                '.item', item_id=search_results[0]['id'], _external=True
            ) if search_results[0] else '',
            back_url=criteria.build_url(page_num=list_page_num),
            time=time.process_time() - start_time,
            **context
        )

    if total_count > 0:
        for i, result in enumerate(search_results):
            result['url'] = criteria.build_url(
                page_num=(criteria.page_num - 1) * (criteria.page_len or 0) + i + 1,
                page_len=1
            )
        if context['is_searching']:
            context['title'] = ngettext('Result', 'Results', total_count)
        else:
            context['title'] = gettext('Full bibliography')
    else:
        context['title'] = gettext('Your search did not match any resources')
    return render_template(
        'kerko/search.html.jinja2',
        form=form,
        search_results=search_results,
        print_url=criteria.build_url(page_len='all', print_preview=True),
        print_preview=criteria.print_preview,
        time=time.process_time() - start_time,
        **context
    )


@blueprint.route('/<string:item_id>')
def item(item_id):
    start_time = time.process_time()

    if current_app.config['KERKO_USE_TRANSLATIONS']:
        babel_domain.as_default()
    result = run_query_unique('id', item_id)

    if not result:
        abort(404)

    build_creators_display(result)
    build_fake_facet_results(result)
    return render_template(
        'kerko/item.html.jinja2',
        item=result,
        title=result.get('data', {}).get('title', ''),
        item_url=url_for('.item', item_id=result['id'], _external=True) if result else '',
        time=time.process_time() - start_time,
        locale=get_locale(),
    )
